(* $Id: lapack_enabled.ml,v 1.3 2004/04/09 01:52:55 ohl Exp $ *)
(* Copyright (C) 2002-2004 by Thorsten Ohl <ohl.uni-wuerzburg.de> et al.
   O'Mega is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by 
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   O'Mega is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  *)  

open Bigarray

type vector = (float, float64_elt, fortran_layout) Array1.t
type matrix = (float, float64_elt, fortran_layout) Array2.t
type ivector = (int32, int32_elt, fortran_layout) Array1.t

external eigenv_glue : matrix -> vector -> vector -> int = "eigenv_glue"

(*i
external eigenw_glue : matrix -> vector -> vector -> ivector -> int = "eigenw_glue"
i*)

let nblock = 64

let to_array1 ba = 
  let dim = Array1.dim ba in
  Array.init dim (fun i -> Array1.get ba (succ i))

(*i
let to_array2 ba = 
  let dim1 = Array2.dim1 ba
  and dim2 = Array2.dim2 ba in
  Array.init dim1 (fun i1 ->
    Array.init dim2 (fun i2 -> Array2.get ba (succ i1) (succ i2)))

let to_array2_transposed ba = 
  let dim1 = Array2.dim2 ba
  and dim2 = Array2.dim1 ba in
  Array.init dim1 (fun i1 ->
    Array.init dim2 (fun i2 -> Array2.get ba (succ i2) (succ i1)))
i*)

exception Input_matrix
exception Input_matrix_size
exception No_convergence of int
exception Internal of string
exception Not_available

let internal name =
  raise (Internal name)

let diagonalize_in_place a =
  assert (Array2.dim1 a = Array2.dim2 a);
  let n = Array2.dim1 a in
  let w = Array1.create float64 fortran_layout n in
  let lwork = (nblock + 2) * n in
  let work = Array1.create float64 fortran_layout lwork in
  let info = eigenv_glue a w work in
  if info = 0 then
    (w, a)
  else if info > 0 then
    raise (No_convergence info)
  else if info = -1 then
    internal "job"
  else if info = -2 then
    internal "upper/lower triangular"
  else if info = -3 then
    raise Input_matrix_size
  else if info = -4 then
    raise Input_matrix
  else if info = -5 then
    internal "leading dimension ( = order of input matrix)"
  else if info = -6 then
    internal "container for eigenvalues"
  else if info = -7 then
    internal "work space"
  else if info = -8 then
    internal "size of work space"
  else if info = -9 then
    internal "container for return value"
  else
    internal "CAN'T HAPPEN!!!"

let clone a =
  let a' =
    Array2.create float64 fortran_layout (Array2.dim1 a) (Array2.dim2 a) in
  Array2.blit a a';
  a'

let diagonalize_big a =
  let n = Array2.dim1 a in
  let eval, evec = diagonalize_in_place (clone a) in
  Array.init n (fun i ->
    (Array1.get eval (succ i), Array2.slice_right evec (succ i)))

let diagonalize' a =
  Array.map (fun (eval, evec) -> (eval, to_array1 evec))
    (diagonalize_big (Array2.of_array float64 fortran_layout a))

let diagonalize a =
  let n = Array.length a in
  let eval, evec =
    diagonalize_in_place (Array2.of_array float64 fortran_layout a) in
  Array.init n (fun i ->
    (Array1.get eval (succ i),
     Array.init n (fun j -> Array2.get evec (succ j) (succ i))))

(*i
let transpose a =
  Array.init (Array.length a.(0))
    (fun i -> Array.init (Array.length a) (fun j -> a.(j).(i)))
i*)

let apply a v =
  assert (Array.length a.(0) = Array.length v);
  let av = Array.make (Array.length a) 0.0 in
  for i = 0 to pred (Array.length a) do
    for j = 0 to pred (Array.length v) do
      av.(i) <- av.(i) +. a.(i).(j) *. v.(j)
    done
  done;
  av

let scale x v =
  Array.init (Array.length v) (fun i -> x *. v.(i))

let random_symmetric n =
  let a = Array.make_matrix n n 0.0 in
  for i = 0 to pred n do
    for j = i to pred n do
      a.(i).(j) <- Random.float 1.0;
      a.(j).(i) <- a.(i).(j)
    done;
  done;
  a

let test n =
  let a = random_symmetric n in
  Array.mapi (fun i (eval, evec) -> (apply a evec, scale eval evec))
    (diagonalize a)

let test' n =
  let a = random_symmetric n in
  Array.mapi (fun i (eval, evec) -> (apply a evec, scale eval evec))
    (diagonalize' a)

(* \begin{dubious}
     Note that \texttt{libcamlrun} must be linked \emph{before}
     \texttt{libg2c.a}:
\begin{verbatim}
  ocamlmktop -custom bigarray.cma \
    lapack.mli lapack.ml lapack_f77.o lapack_glue.o \
    -cclib "-lcamlrun -llapack -lblas -lg2c"
\end{verbatim}
   \end{dubious} *)

(*i
 *  Local Variables:
 *  mode:caml
 *  indent-tabs-mode:nil
 *  page-delimiter:"^(\\* .*\n"
 *  End:
i*)
