(* $Id: vertex_syntax.ml,v 1.6 2003/07/14 17:19:41 ohl Exp $ *)
(* Copyright (C) 2003 by Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
   O'Mega is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by 
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   O'Mega is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  *)  

(* \thocwmodulesection{Abstract Syntax} *)

type index = int

(* \begin{dubious}
     \emph{The following is not complete yet.}
     We would like to allow scalars as coefficients of vectors.  Since
     recursive functors are not available in O'Caml yet, we will have to
     go back to a polymorphic implementation.
  \end{dubious} *)

module R = Algebra.Make_Ring(Algebra.Small_Rational)(Algebra.Term)
module V = Algebra.Make_Linear(R)

type current = S | P | SL | SR | V | A | VL | VR

let current_to_string = function
  | S -> "S"
  | P -> "P"
  | SL -> "S-P"
  | SR -> "S+P"
  | V -> "V"
  | A -> "A"
  | VL -> "V-A"
  | VR -> "V+A"

type vector =
  | Polarization of index (* $\epsilon_n^{\mu_n}$ *)
  | Momentum of index (* $k_n^{\mu_n}$ *)
  | Current of current * index * index (* $\bar\psi_n\Gamma\psi_m$ *)
  | External of string
  | Pseudo of vector * vector * vector (* $\epsilon^{\mu\nu\rho\sigma}v_\nu v_\rho v_\sigma$ *)
  | Sum of (vector, scalar) V.t

and scalar = 
  | Constant of string
  | Dot of vector * vector
  | Eps of vector * vector * vector * vector

type scalar_expr = scalar R.t
let null = R.null
let integer i = R.scale (R.C.make i 1) (R.unit ())
let fraction x i = R.scale (R.C.make 1 i) x
let multiple i x = R.scale (R.C.make i 1) x
let mul = R.mul
let add = R.add
let sub = R.sub

type vector_expr = (vector, scalar) V.t

let rec scalar_to_string = function
  | Constant s -> s
  | Dot (v1, v2) ->
      "(" ^ vector_to_string v1 ^ "." ^ vector_to_string v2 ^ ")"
  | Eps (v1, v2, v3, v4) ->
      "eps(" ^ vector_to_string v1 ^ "," ^ vector_to_string v2  ^ "," ^
      vector_to_string v3 ^ "," ^ vector_to_string v4 ^ ")"

and vector_to_string = function 
  | Polarization n -> "e" ^ string_of_int n
  | Momentum n -> "k" ^ string_of_int n
  | Current (c, n, m) ->
      Printf.sprintf "<%d|%s|%d>" n (current_to_string c) m
  | External p -> "<<" ^ p ^ ">>"
  | Pseudo (v1, v2, v3) ->
      "eps(" ^ vector_to_string v1 ^ "," ^ vector_to_string v2  ^ "," ^
      vector_to_string v3 ^ ")"
  | Sum vs -> V.to_string vector_to_string scalar_to_string vs

let scalar_expr_to_string = R.to_string scalar_to_string
let vector_expr_to_string = V.to_string vector_to_string scalar_to_string

let derive_scalar v = function
  | Constant _ -> None
  | Dot (v', v1) when v' = v -> Some v1
  | Dot (v1, v') when v' = v -> Some v1
  | Dot (_, _) -> None
  | Eps (v', v1, v2, v3) when v' = v -> Some (Pseudo (v1, v2, v3))
  | Eps (v1, v', v3, v2) when v' = v -> Some (Pseudo (v1, v2, v3))
  | Eps (v2, v3, v', v1) when v' = v -> Some (Pseudo (v1, v2, v3))
  | Eps (v3, v2, v1, v') when v' = v -> Some (Pseudo (v1, v2, v3))
  | Eps (_, _, _, _) -> None

let e i = Polarization i
let k i = Momentum i

let dot v1 v2 = 
  R.atom (if v1 <= v2 then Dot (v1, v2) else Dot (v2, v1))

let eps v1 v2 v3 v4 = 
  R.atom (Eps (v1, v2, v3, v4))

let pseudo v1 v2 v3 = 
  Pseudo (v1, v2, v3)

let addv v1 v2 = 
  Sum (V.add (V.atom v1) (V.atom v2))

let subv v1 v2 = 
  Sum (V.sub (V.atom v1) (V.atom v2))

let current c i j =
  Current (c, i, j)

let constant s =
  R.atom (Constant s)

let partial v s =
  V.partial (derive_scalar v) s

(*i
let scalev c v = 
  Sum (V.scale (V.C.atom c) (V.atom v))
i*)

exception Syntax_Error of string * int * int

(*i
 *  Local Variables:
 *  mode:caml
 *  indent-tabs-mode:nil
 *  page-delimiter:"^(\\* .*\n"
 *  compile-command:"ocamlc -o vertex thoList.ml{i,} pmap.ml{i,} vertex.ml"
 *  End:
i*)
